// src/routes/mqtt/+server.js
import { db } from "$lib/server/db";
import * as table from "$lib/server/db/schema";
import { connectedDevices, deviceSensorData, getCurrentDevices } from "$lib/server/mqtt-devices.js";
import { eq } from "drizzle-orm";
import mqtt from "mqtt";
import { writable } from "svelte/store";

// A Svelte store to hold the latest MQTT message.
// In a real application, you might want to store more data or
// use a more robust way to manage messages, but for a basic example, this works.
const latestMessage = writable("No message yet");
const devices = writable([]);

let client = null;

// Function to update device status and sensor data
function updateDevice(deviceId, sensorData = null) {
  const now = new Date();
  const deviceName = `ESP8266 #${deviceId.slice(-6)}`;

  // Update device info
  connectedDevices.set(deviceId, {
    id: deviceId,
    name: deviceName,
    type: "esp8266",
    status: "online",
    lastSeen: now,
  });

  // Update sensor data if provided
  if (sensorData) {
    deviceSensorData.set(deviceId, {
      ...sensorData,
      timestamp: now,
    });
  }

  // Update the devices store
  updateDevicesStore();
}

// Function to parse sensor data from MQTT payload
function parseSensorData(payload) {
  try {
    const data = JSON.parse(payload);
    return {
      temperature: parseFloat(data.temperature) || 0,
      humidity: parseFloat(data.humidity?.replace("%", "")) || 0,
      pressure: parseFloat(data.pressure) || 0,
      altitude: parseFloat(data.altitude) || 0,
    };
  } catch (error) {
    console.error("Error parsing sensor data:", error);
    return null;
  }
}

// Store SSE controllers for cleanup
const sseControllers = new Set();

// Function to update devices store with latest data
function updateDevicesStore() {
  const deviceList = Array.from(connectedDevices.values()).map((device) => {
    const sensorData = deviceSensorData.get(device.id);
    return {
      ...device,
      sensorData: sensorData || null,
    };
  });
  devices.set(deviceList);
}

// Clean up offline devices (not seen in last 5 minutes)
function cleanupDevices() {
  const fiveMinutesAgo = new Date(Date.now() - 5 * 60 * 1000);
  let updated = false;

  for (const [deviceId, device] of connectedDevices.entries()) {
    if (device.lastSeen < fiveMinutesAgo && device.status === "online") {
      connectedDevices.set(deviceId, { ...device, status: "offline" });
      updated = true;
    }
  }

  if (updated) {
    updateDevicesStore();
  }
}

// Run cleanup every minute
setInterval(cleanupDevices, 60000);

// getCurrentDevices is now imported from the shared module

// Function to connect to MQTT
function connectMqtt() {
  if (client && client.connected) {
    return; // Already connected
  }

  // Replace with your MQTT broker details
  const BROKER_URL = "mqtt://kada49.it:1883"; // Example public broker
  //const BROKER_URL = "mqtt://test.mosquitto.org:1883";
  const TOPIC = "esp8266/+/data"; // Replace with your desired topic

  client = mqtt.connect(BROKER_URL);

  client.on("connect", () => {
    console.log("Connected to MQTT broker!");
    client.subscribe(TOPIC, (err) => {
      if (!err) {
        console.log(`Subscribed to topic: ${TOPIC}`);
      } else {
        console.error(`Subscription error: ${err}`);
      }
    });
  });

  client.on("message", async (topic, message) => {
    const payload = message.toString();
    console.log(`Received message from topic "${topic}": ${payload}`);
    latestMessage.set(payload); // Update the Svelte store

    // Extract device ID from topic (esp8266/DEVICE_ID/data)
    const topicParts = topic.split("/");
    if (topicParts.length >= 3 && topicParts[0] === "esp8266" && topicParts[2] === "data") {
      const deviceId = topicParts[1];
      console.log(`Processing data for device: ${deviceId}`);

      // Parse sensor data
      const sensorData = parseSensorData(payload);
      if (sensorData) {
        console.log(`Parsed sensor data:`, sensorData);
        updateDevice(deviceId, sensorData);
        const devices = await db.select().from(table.sensors).where(eq(table.sensors.id, deviceId));
        if (devices.length == 1)
          await db
            .insert(table.sensorData)
            .values({
              sensor: deviceId,
              temperature: sensorData.temperature,
              humidity: sensorData.humidity,
              altitude: sensorData.altitude,
              pressure: sensorData.pressure,
            });
      } else {
        // Still update device as online even if data parsing failed
        updateDevice(deviceId);
      }
    }
  });

  client.on("error", (err) => {
    console.error(`MQTT error: ${err}`);
    if (client) {
      client.end(); // Close connection on error
    }
    client = null;
    // Implement re-connection logic here if needed
  });

  client.on("close", () => {
    console.log("MQTT connection closed.");
    client = null;
    // Implement re-connection logic here if needed
  });
}

// Connect to MQTT when the server starts
connectMqtt();

/** @type {import("./$types").RequestHandler} */
export async function GET({ request }) {
  const headers = {
    "Content-Type": "text/event-stream",
    "Cache-Control": "no-cache",
    Connection: "keep-alive",
  };

  return new Response(
    new ReadableStream({
      start(controller) {
        console.log("Client connected to MQTT SSE stream");

        let isConnected = true;
        let interval;

        // Add this controller to the broadcast set
        sseControllers.add(controller);

        // Cleanup function
        const cleanup = () => {
          console.log("Cleaning up MQTT SSE connection");
          isConnected = false;
          if (interval) {
            clearInterval(interval);
            interval = null;
          }
          sseControllers.delete(controller);
          if (unsubscribe) {
            unsubscribe();
          }
        };

        // Send current device data
        function sendDevices() {
          if (!isConnected) {
            cleanup();
            return;
          }

          try {
            const devices = getCurrentDevices();
            const message = {
              devices: devices,
              timestamp: new Date().toISOString(),
            };

            // Check if controller is still valid before enqueueing
            if (isConnected) {
              controller.enqueue(`data: ${JSON.stringify(message)}\n\n`);
            }
          } catch (error) {
            // Handle controller closed error specifically
            if (
              error.code === "ERR_INVALID_STATE" ||
              error.message.includes("Controller is already closed")
            ) {
              console.log("MQTT SSE controller closed, stopping interval");
            } else {
              console.error("Error sending MQTT device data:", error);
            }
            cleanup();
          }
        }

        // Send initial data
        sendDevices();

        // Send updates every 2 seconds
        interval = setInterval(sendDevices, 2000);

        // Subscribe to MQTT messages via the store for real-time updates
        const unsubscribe = latestMessage.subscribe((message) => {
          if (message !== "No message yet" && isConnected) {
            try {
              const data = `data: ${JSON.stringify({
                message: message,
                devices: getCurrentDevices(),
                timestamp: new Date().toISOString(),
              })}\n\n`;
              if (isConnected) {
                controller.enqueue(data);
              }
            } catch (error) {
              if (
                error.code === "ERR_INVALID_STATE" ||
                error.message.includes("Controller is already closed")
              ) {
                console.log("MQTT SSE controller closed during message send");
                cleanup();
              }
            }
          }
        });

        // Handle client disconnection
        request.signal.addEventListener("abort", () => {
          console.log("Client disconnected from MQTT SSE stream");
          cleanup();
        });
      },
      cancel() {
        console.log("MQTT SSE stream cancelled");
      },
    }),
    { headers },
  );
}
